# -*- coding: UTF-8 -*-
###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# libsecure.py
#
# classes utilitaires pour lancement des services en https
# et vérification des certificats
#
###########################################################################

import sys, os
from glob import glob
import httplib, xmlrpclib
from creole.parsedico import parse_dico
from creole import config
from twisted.python import log

# SSL imports
from OpenSSL import SSL
import sys, xmlrpclib
from time import time
from urlparse import urlparse

try:
    # Get C binding version
    from OpenSSL.crypto import X509_verify_cert_error_string
except ImportError:
    try:
        # ctypes binding version
        from ctypes import cdll, c_ulong, c_long, c_char_p, c_int, create_string_buffer
        from ctypes.util import find_library

        _libssl = cdll.LoadLibrary(find_library('ssl'))

        X509_verify_cert_error_string = _libssl.X509_verify_cert_error_string
        X509_verify_cert_error_string.restype = c_char_p
        X509_verify_cert_error_string.argtypes = (c_long,)

    except (ImportError, AttributeError, TypeError):
        # ctypes module or X509_verify_cert_error_string symbol is missing:
        # use dummy fallback
        def X509_verify_cert_error_string(errnum):
            return str(errnum)

# transport sécurisé utilisant un certificat
class TransportEole(xmlrpclib.SafeTransport):

    def __init__(self):
        dico = parse_dico()
        if dico.has_key('sso_cert') and os.path.isfile(dico['sso_cert']):
            self.certfile = dico['sso_cert']
        else:
            self.certfile = config.cert_file
        super(xmlrpclib.SafeTransport, self).__init__()

    def make_connection(self, host):
        # create a HTTPS connection object from a host descriptor
        # host may be a string, or a (host, x509-dict) tuple
        localhost, extra_headers, x509 = self.get_host_info(host)
        try:
            HTTPS = httplib.HTTPS
        except AttributeError:
            raise NotImplementedError(
                "your version of httplib doesn't support HTTPS"
                )
        else:
            cx = HTTPS(localhost, None,
                       key_file = self.certfile,
                       cert_file = self.certfile)
            return cx

class ClientContextFactory:
    """context factory for SSL clients."""
    isClient = 1
    method = SSL.SSLv3_METHOD

    def __init__(self, certfile='', ca_location='', mode=None):
        if certfile != '':
            self.certfile = certfile
        else:
            self.certfile = config.cert_file
        self.ca_location = ca_location
        if mode is None:
            self.verify_mode = SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT
        else:
            self.verify_mode = mode
        self.peer_cert_data = {}

    # fonction de vérification d'un certificat (utilisé pour les proxies)
    def _verify(self, connection, x509, errnum, errdepth, ok):
        """ This function will be called once for each certificate in the chain
        """
        self.peer_cert_data['subject'] = x509.get_subject()
        self.peer_cert_data['issuer'] = x509.get_issuer()
        self.peer_cert_data['expired'] = x509.has_expired()
        self.peer_cert_data['error'] = errnum
        print '  error %s at depth %s lookup: %s' % (
            errnum, errdepth, X509_verify_cert_error_string(errnum))
        return ok

    def getContext(self):
        ctx = SSL.Context(self.method)
        ctx.use_certificate_file(self.certfile)
        try:
            ctx.use_privatekey_file(self.certfile)
        except:
            # impossible de charger la clé depuis le fichier du certificat, on regarde si il n'y a pas un fichier .key associé
            alt_keyfile = os.path.splitext(self.certfile)[0] + '.key'
            ctx.use_privatekey_file(alt_keyfile)
        # chargement des certificats des autorités de certification (si disponible)
        ca_certs = glob(self.ca_location)
        for ca_cert in ca_certs:
            try:
                ctx.load_verify_locations(ca_cert)
            except SSL.Error, e:
                log.msg('Erreur au chargement du certificat %s : %s' % (ca_cert, str(e)))
        ctx.set_verify(self.verify_mode, self._verify)
        ctx.set_verify_depth(10)
        return ctx

class ServerContextFactory:
    """
    Factory permettant de créer un contexte SSL
    """

    def __init__(self,certfile=''):
        if certfile != '':
            self.certfile = certfile
        else:
            self.certfile = config.cert_file

    def getContext(self):
        """Create an SSL context.
        This is a sample implementation that loads a certificate from a file
        called 'server.pem'."""
        ctx = SSL.Context(SSL.SSLv23_METHOD)
        ctx.use_certificate_file(self.certfile)
        try:
            ctx.use_privatekey_file(self.certfile)
        except:
            # impossible de charger la clé depuis le fichier du certificat, on regarde si il n'y a pas un fichier .key associé
            alt_keyfile = os.path.splitext(self.certfile)[0] + '.key'
            ctx.use_privatekey_file(alt_keyfile)
        return ctx

