#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authform.py
#
# service web de connexion au service sso Eole
#
# compatible avec les versions 1 et 2 de la librairie CAS
# (http://www.ja-sig.org/products/cas/)
#
###########################################################################

import sys, xmlrpclib, os
import time
import urlparse
from urllib2 import urlopen, quote, unquote
from cgi import escape
from OpenSSL import SSL
from socket import error as socket_err

from twisted.internet import reactor, ssl, protocol, defer
from twisted.web2 import static, server, channel, http, responsecode
from twisted.web2.resource import PostableResource as Resource
from twisted.web2.http_headers import Cookie as TwCookie, MimeType
from twisted.web2.xmlrpc import XMLRPC
from twisted.web.client import getPage

from creole.cert import gen_certif
# Imports EoleSSO
from config import *
from eolesso.util import *
from eolesso.libsecure import ServerContextFactory, ClientContextFactory
from eolesso.dataproxy import LDAPProxy
from eolesso import ticketcache
from eolesso.errors import (Redirect, CasError, MissingParameters, InvalidTicket,
                   InternalError)
from authserver import SSOSessionManager, log

# OPENID
from oidprovider import OIDRoot
# SAML
import saml_idp

# initialisations pour la localisation
APP = 'eolesso'
import locale, gettext
gettext.bindtextdomain (APP)
gettext.textdomain (APP)
# installons la fonction _() dans les builtins
gettext.install (APP, unicode=0)

try:
    _ = _
except NameError:
    _ = unicode

class XMLRPCManager(XMLRPC):

    def __init__(self,session_manager):
        self.manager = session_manager
        XMLRPC.__init__(self)

    def xmlrpc_authenticate(self, post_data, username, password):
        log.msg('* xmlrpc authenticate')
        return self.manager.authenticate(username, password)

    def xmlrpc_validate_session(self, post_data, session_id):
        log.msg('* xmlrpc validate_session')
        return self.manager.validate_session(session_id)

    def xmlrpc_verify_session_id(self, post_data, session_id):
        log.msg('* xmlrpc verify_session_id')
        return self.manager.verify_session_id(session_id)

    def xmlrpc_logout(self, post_data, session_id):
        log.msg('* xmlrpc logout')
        return self.manager.logout(session_id)

    def xmlrpc_verify_app_ticket(self, post_data, app_ticket, appurl):
        log.msg('* xmlrpc verify_app_ticket')
        return self.manager.verify_app_ticket(app_ticket, appurl)

    def xmlrpc_get_user_details(self, post_data, app_ticket, appurl, sections=False, renew=False):
        log.msg('* xmlrpc get_user_details')
        return self.manager.get_user_details(app_ticket, appurl, sections, renew)

    def xmlrpc_get_user_info(self, post_data, session_id, details=False):
        log.msg('* xmlrpc get_user_infos')
        return self.manager.get_user_info(session_id, details)


class CasResource(Resource):
    content_type = 'text/html'
    required_parameters = ()

    def render(self, request):
        headers = {'Content-type': MimeType.fromString(self.content_type)}
        errheaders = {'Content-type': MimeType.fromString('application/xml')}
        try:
            self.check_required_parameters(request)
            return self._render(request)
        except Redirect, exc:
            location = exc.location
            headers['location'] =  location
            log.msg('redirecting to %s', location)
            # 303 See other
            return http.Response(code=303, headers=headers)
        except CasError, exc:
            return http.Response(stream=exc.as_xml(), headers=errheaders)
        except Exception, exc: # unknown error
            error = InternalError(str(exc))
            return http.Response(stream=error.as_xml(), headers=errheaders)

    def check_required_parameters(self, request):
        """raises a `MissingParameters` exception if all required parameters
        are not found in `request.args`.
        """
        missing = [param for param in self.required_parameters
                   if param not in request.args]
        if missing:
            raise MissingParameters(missing)

class AuthenticationVerify(CasResource):
    """ Page de vérification de l'identité
        démarre une session et stocke un cookie dans le client
        si les identifiants sont corrects"""
    isLeaf = False
    addSlash = False
    required_parameters = ('username','password')

    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def _render(self, request):
        try:
            username = unicode(request.args['username'][0],encoding)
        except:
            return http.Response(stream=gen_page("Authentification","Accès non autorisé !"))
        try:
            password = unicode(request.args['password'][0],encoding)
        except:
            password = ""
        try:
            from_url = request.args['service'][0]
        except:
            from_url = ""
        try:
            redirect_to = request.args['redirect_to'][0]
        except KeyError:
            redirect_to = ''
        # vérification du login ticket
        try:
            pass
            #ticket = request.args['lt'][0]
            #assert self.manager.login_sessions.validate_session(ticket)
        except:
            # on n'est pas passé par /login : on renvoie dessus
            if from_url:
                response = http.RedirectResponse('/login?service=%s' % from_url)
            else:
                response = http.RedirectResponse('/login')
            return response
        # on va tenter l'authentification sur tous les serveurs d'auth
        session_id = self.manager.authenticate(username, password)
        # authentification distante (SSO) reussie
        if session_id != "":
            # on stocke un cookie dans le navigateur
            # chaine = '%s,%s'%(session_id), username.encode(encoding))
            session_id = session_id.encode(encoding)
            cookies = getCookies(request)
            for cookie in cookies:
                if cookie.name == 'EoleSSOServer':
                    cookies.remove(cookie)
            cookies.append(TwCookie("EoleSSOServer", session_id, path = "/", discard=True, secure=True))
            if from_url == '':
                # pas de service spécifié, on indique que la session est ouverte, ou on redirige si demandé
                if redirect_to != '':
                    target_url = redirect_to
                else:
                    target_url = '%s/loggedin' % (AUTH_FORM_URL)
            else:
                # on redirige sur la page de l'application de destination
                app_ticket = self.manager.get_app_ticket(session_id, from_url, from_credentials=True)
                target_url = urljoin(from_url, 'ticket=%s' % app_ticket)
                log.msg("redirecting to calling app with ticket : %s" % from_url)
            if self.manager.user_sessions.get(session_id,'') == None:
                # Si session délivrée par le parent, rediriger sur lui pour poser un cookie d'auth qu'il puisse lire
                response = http.RedirectResponse('%s/gen_cookie?ticket=%s&return_url=%s' % (self.manager.parent_url, quote(session_id), quote(target_url)))
                log.msg("redirecting to parent server for session setting")
                log.msg('%s/gen_cookie?ticket=%s&return_url=%s' % (self.manager.parent_url, quote(session_id), quote(target_url)))
            else:
                response = http.RedirectResponse(target_url)
            if cookies:
                response.headers.setHeader('Set-Cookie', cookies)
            return response

        # aucun des serveurs n'a authentifié l'utilisateur (meme l'authentification locale), on revient
        # sur la page de saisie de username/password
        return http.RedirectResponse('/?service=%s&failed=1' %(from_url))

class CASCompliantResponse(CasResource):
    """implémentation du protocole CAS pour la validation de ticket
    """

    authorized_codes=[responsecode.OK,
                      responsecode.MULTIPLE_CHOICE,
                      responsecode.MOVED_PERMANENTLY,
                      responsecode.FOUND,
                      responsecode.SEE_OTHER,
                      responsecode.NOT_MODIFIED,
                      responsecode.USE_PROXY,
                      responsecode.TEMPORARY_REDIRECT,
                      ]

    required_parameters = ('service','ticket')

    def __init__(self, manager, serve_proxy = False):
        Resource.__init__(self)
        self.manager = manager
        self.serve_proxy = serve_proxy

    def _gen_user(self, infos, data_filter):
        """génération de la partie données utilisateur de la réponse CAS
        """
        data = ""
        if data_filter != "":
            for section, attrs in data_filter.items():
                data += "    <cas:%s>\n%s    </cas:%s>\n" % (section, self._gen_section(attrs, infos), section)
        else:
            data = "    <cas:user>%s</cas:user>" % infos['uid'][0]
        return data

    def _gen_section(self,attrs,infos):
        """génère une section dans la réponse
        """
        section = ""
        for name, datakey in attrs.items():
            for val in infos[datakey]:
                section += "      <cas:%s>%s</cas:%s>\n" % (name, val, name)
        return section

    def check_proxy_callback(self, cb_url, orig_parameters):
        """vérifie la validité d'une url de callback founie par un service proxy:
        - l'url doit être en https
        - le certificat SSL doit être valide
        - le nom du certificat doit correspondre au service
        """
        url = urlparse.urlparse(cb_url)
        if url.scheme == 'https':
            # validité du certificat
            ctx = ClientContextFactory(mode=SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT, ca_location=CA_LOCATION)
            # connexion à l'url de callback pour vérifier le certificat (port 443 si pas de port spécifié)
            d = getPage(cb_url, contextFactory=ctx, timeout=10)
            return d.addCallbacks(self.proxy_cert_ok, self.proxy_errb, callbackArgs=[ctx, orig_parameters, url], errbackArgs=[orig_parameters,url])
        # url de proxy non valide
        log.msg("!! url de proxy non valide (https obligatoire) !!")
        return self.validate_st(orig_parameters)

    def proxy_cert_ok(self, stream, ctx, orig_parameters, url):
        """gestion de la demande de PGT si proxy valide
        - envoi d'une requête GET sur l'url de callback du proxy avec les paramètres pgtId et pgtIou
        - vérification du retour : OK si code 200 (ou 3xx si redirection)
        - la validation du ticket d'origine (ST ou PT) continue, avec ajout de pgtIou si vérification OK
        """
        service = orig_parameters[2]
        ticket = orig_parameters[3]
        ctx.peer_cert_data['issuer'] = service # HACK !!!!
        # vérification de la concordance service / certificat XXX FIXME : issuer ou subject ???
        if ctx.peer_cert_data['issuer'] != service:
            # certificat non valide
            log.msg("!! vérification du certificat : le nom du certificat ne correspond pas au service !!")
        else:
            pgt_iou = None
            # Création d'un pgt temporaire (et pgtiou associé) en attendant la validation ?
            app_ticket = self.manager.get_proxy_granting_ticket(ticket, service, url)
            if app_ticket is not None:
                # préparation de la requête à envoyer sur le callback du proxy
                u = urlopen("%s?pgtId=%s&pgtIou=%s" % (url.geturl(), app_ticket.pgt, app_ticket.pgtiou))
                data = u.read()
                if u.code in self.authorized_codes:
                    log.msg("** proxy response code OK : %s" % u.code)
                    # le proxy doit retourner un code 200 ou 3xx après envoi du pgtId et pg_tiou
                    return self.validate_st(orig_parameters, app_ticket)
                # réponse du proxy invalide, on n'envoie pas de pgt iou à la validation du ticket
                log.msg("** proxy response code ERROR : %s" % u.code)
            else:
                log.msg("!! obtention d'un ticket PGT echouée pour le service !!")
        return self.validate_st(orig_parameters)

    def proxy_errb(self, err, orig_parameters, url):
        log.msg("!! proxy non validé : %s !!" % url.geturl())
        log.msg("\n!! --> %s !!\n" % err.getErrorMessage())
        return self.validate_st(orig_parameters)

    def _render(self, request):
        error = {'code':'INTERNAL_ERROR','detail':''}
        data = error, None
        ok = False

        from_url = request.args['service'][0]
        ticket = request.args['ticket'][0]
        if not self.serve_proxy and ticket.startswith('PT'):
            raise InvalidTicket(ticket,
                                'received a PT ticket instead of a ST')
        # Récupération du paramètre renew si présent
        if request.args.get('renew', None) is not None:
            renew = True
        else:
            renew = False
        # CAS 2 : récupération de l'url de callback pour les proxys (pgtURl : optionnel)
        pgt_url = None
        if CAS_VERSION == 2:
            try:
                pgt_url = request.args['pgtUrl'][0]
                return self.check_proxy_callback(pgt_url, [ok, data, from_url, ticket, renew])
            except KeyError:
                # Cas sans pgtUrl, on continue la validation du ticket normalement
                pass
        return self.validate_st([ok, data, from_url, ticket, renew])

    def _traversed_proxies(self, ticket):
        """return the <cas:proxies> node listing traversed proxies

        NOTE: /serviceValidate and /proxyValidate could share the same
        implementation since ticket.proxypath() should return an empty
        list in case of ST apptickets, but just in case, we return
        an hard-coded empty node for ST tickets
        """
        proxies = '\n'.join('    <cas:proxy>%s</cas:proxy>' % url
                            for url in ticket.proxypath())
        return u'\n  <cas:proxies>\n%s\n  </cas:proxies>' % proxies

    def validate_st(self, validate_parameters, pgt=None):
        ok, data, from_url, ticket, renew = validate_parameters
        if ticket is not None:
            if pgt is not None:
                pgt_data = "\n  <cas:proxyGrantingTicket>%s</cas:proxyGrantingTicket>" % pgt.pgtiou
                # on renvoie les proxys traversés si besoin
                if pgt.proxypath() != []:
                   pgt_data += self._traversed_proxies(pgt)
            else:
                # si on n'a pas de paramètre pgt_iou, on considère que d'éventuels tests
                # sur une url de proxy ont échoué
                self.manager.invalidate_proxy(ticket)
                pgt_data = ""
            # si ticket inconnu du serveur, on va tenter sa validation sur tous les serveurs d'auth
            ok, data = self.manager.get_user_details(ticket, from_url, True, renew)
        infos, filter_data = data
        if ok:
            # authentification distante (SSO) reussie
            # génération des informations utilisateur
            user_infos = self._gen_user(infos, filter_data)
            # formatting response
            if CAS_VERSION == 2:
                resp_body = user_infos + pgt_data
                response = cas_response('authenticationSuccess', resp_body)
            else:
                response = 'yes\n%s\r\n\r\n' % infos['uid'][0]
        else:
            error, filter_data  = data
            # aucun des serveurs n'a authentifié l'utilisateur (meme l'authentification locale), on revient
            # sur la page de saisie de username/password
            if CAS_VERSION == 2:
                response = cas_response('authenticationFailure', escape(error['detail']), code=error['code'])
            else:
                response = 'no\r\n\r\n'

        cas_resp = http.Response(stream=response)
        cas_resp.headers.setRawHeaders('Pragma', 'no-cache')
        cas_resp.headers.setRawHeaders('Cache-Control', 'no-store')
        cas_resp.headers.setRawHeaders('Expires', 'Mon, 26 Jul 1997 05:00:00 GMT')
        return cas_resp

class ProxyResource(CasResource):
    """
    page de génération d'un ticket de type PT
    """
    addSlash = False

    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def _render(self, request):
        try:
            pgt = request.args['pgt'][0]
            target_service = request.args['targetService'][0]
        except:
            error = {'code':'INVALID_REQUEST','detail':'paramètre manquant'}
        # vérification du pgt
        if pgt in self.manager.proxy_granting_sessions:
            # génération du ticket PT
            ticket = self.manager.get_proxy_ticket(pgt, target_service)
            response_body = "<cas:proxyTicket>%s</cas:proxyTicket>" % ticket
            response = cas_response('proxySuccess', response_body)
            return http.Response(stream=response)
        else:
            error = {'code':'BAD_PGT', 'detail':"le pgt fourni n'est pas reconnu par le service : %s" % pgt}
        response = cas_response('proxyFailure', escape(error['detail']), code=error['code'])
        return http.Response(stream=response)


class LoggedInForm(CasResource):
    """
    page de confirmation d'ouverture de session (si pas d'url de redirection fournie)
    """
    addSlash = False

    def _render(self,request):
        content = """<form action="/logout" method="post"><input type="Submit" value="Fermer la connexion"></form>"""
        return http.Response(stream=gen_page("Session SSO valide",content))

class ErrorPage(CasResource):
    """
    page d'erreur http
    """
    addSlash = False

    def __init__(self, code, description):
        self.code = code
        self.description = description
        title = ('Eole-SSO : Erreur')
        super(ErrorPage, self).__init__()

    def _render(self,request):
        return http.StatusResponse(code=self.code, description=self.description)

class LogoutForm(CasResource):
    """
    page de confirmation de logout (si pas d'url de redirection fournie)
    """
    addSlash = False

    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def _render(self,request):
        # invalidation de la session sur le serveur d'authentification
        cookies = getCookies(request)
        for cookie in cookies:
            if cookie.name == 'EoleSSOServer':
                session_id = cookie.value
                if session_id is not None:
                    # suppression de la session sur le manager
                    self.manager.logout(session_id.encode(encoding))
                    cookie.expires = 1
                break
        if request.args.has_key('service'):
            url_from = request.args['service'][0]
            response = http.RedirectResponse(url_from)
            log.msg("logged out : redirecting to : %s" % url_from)
        else:
            #si pas de parametre from, on affiche la page de logout en local
            content = """<form action="/login" method="post">
            <input type="Submit" value="Connexion"></form>"""
            response = http.Response(stream=gen_page("Fin de la session SSO", content))
        # mise en place des cookies
        if cookies:
            response.headers.setHeader('Set-Cookie', cookies)
        return response

class LocalCookie(CasResource):
    """
    page de génération d'un cookie d'authentification local
    """
    addSlash = False
    required_parameters = ('return_url','ticket')

    def __init__(self, manager):
        Resource.__init__(self)
        self.manager = manager

    def _render(self, request):
        try:
            ticket = unquote(request.args['ticket'][0])
            return_url = request.args['return_url'][0]
        except:
            error = {'code':'INVALID_REQUEST','detail':'paramètre manquant'}
            log.msg(error)
        log.msg("Local cookie requested")
        # on vérifie que la session est bien existante et valide
        response = http.RedirectResponse(unquote(return_url))
        if self.manager.verify_session_id(ticket):
            ticket = ticket.encode(encoding)
            # XXX FIXME : gérer plusieurs niveaux de parents (redirection sur parent si ticket non local)
            if self.manager.user_sessions.get(ticket,'') == None:
                print "redirect one level higher !"
            cookies = getCookies(request)
            for cookie in cookies:
                if cookie.name == 'EoleSSOServer':
                    cookies.remove(cookie)
            cookies.append(TwCookie("EoleSSOServer", ticket, path = "/", discard=True, secure=True))
            if cookies:
                log.msg("Local cookie sent, redirecting to child server")
                response.headers.setHeader('Set-Cookie', cookies)
        return response

class TestForm(CasResource):
    """The resource that is returned when you are not logged in"""
    addSlash = False

    def __init__(self, manager, auth_server, oidserver, *args):
        self.manager = manager
        self.auth_server = auth_server
        self.oidserver = oidserver
        self.content = ''
        Resource.__init__(self)

    def is_true(self, param):
        return param == "true"

    def render_content(self, request):
        """
        protocole CAS:
         1. service : l'URL d'où on vient
         2. renew=true : on force à se réauthentifier
         3. gateway=true : si pas authentifié, on revient à l'URL
            définie par service sans ticket d'application
        ajout Eole
         4. redirect_to : permet de définir une page ou revenir après authentification
            (utile pour l'authentification depuis des protocoles non CAS)
        """
        self.content = ""
        gateway = self.is_true(request.args.get('gateway', ['false'])[0])
        renew = self.is_true(request.args.get('renew', ['false'])[0])
        warn = self.is_true(request.args.get('warn', ['false'])[0])
        try:
            from_url = request.args['service'][0]
        except KeyError:
            # nécessaire car les ressources ne sont instanciées qu'une
            # seule fois
            from_url = ''
        try:
            redirect_to = request.args['redirect_to'][0]
        except KeyError:
            redirect_to = ''
        #on va verifier si on a un identifiant de session dans le cookie
        user_session = getCookie(request, 'EoleSSOServer')
        # XXX: IE et Opera ne gerent pas les cookies de la meme
        # maniere que Firefox sous firefox, quand un cookie devient
        # invalide (au niveau du temps), il est supprimé sous IE et
        # Opera, ils sont gardés, on doit donc tester si ils ont une
        # valeur spéciale, ici foo
        #si on en a bien un, on le verifie auprès du serveur d'authentification
        if not renew and user_session and user_session != 'foo':
            session_id = user_session.value.encode(encoding)
            verif =  self.manager.verify_session_id(session_id)
            # si l'ID de session est verifie, on génère le ticket applicatif,
            # et on redirige vers l'URL de base
            if verif[0]:
                if from_url:
                    app_ticket = self.manager.get_app_ticket(session_id, from_url)
                    url = urljoin(from_url, 'ticket=%s' % app_ticket)
                    if warn:
                        self.content = str("""<html>%s<body onload="alert('%s'); document.location.href='%s'"></body></html>""" % (head % "leaves", _('You are being authenticated'), url))
                        return None
                    return url
                return '/loggedin'

        if gateway and from_url:
            return urljoin(from_url, 'ticket=')
        self.auth_msg = "Veuillez vous authentifier"
        if request.args.has_key('failed'):
            if request.args['failed'][0] == '1':
                self.auth_msg = "Echec de l'authentification, recommencez"
        # création d'un login ticket jouable une seule fois pour cet essai d'authentification
        login_ticket = self.manager.get_login_ticket()
        if from_url != "":
            service_input = """<input type="hidden" name="service" value="%s"/>""" % (from_url)
        else:
            service_input = ""
        if redirect_to != "":
            redirect_input =  """<input type="hidden" name="redirect_to" value="%s"/>""" % (redirect_to)
        else:
            redirect_input = ""

        content ="""<form action='/verify/' method='post'>
          <input type="hidden" name="lt" value="%s">
            %s
            %s
            <table>
               <tr><td>Login : </td><td><input type='text' name='username'/></td></tr>
               <tr><td>Mot de passe : </td><td><input type='password' name='password'/></td></tr>
            </table>
          <input type='submit'/>
        </form>
        """ % (login_ticket, service_input, redirect_input)
        self.content = gen_page(self.auth_msg, content) 
        return None

    def _render(self,request):
        redirect_to = self.render_content(request)
        if redirect_to == None:
            return http.Response(stream=self.content)
        else:
            log.msg("authform, redirecting to : %s" % redirect_to)
            return http.RedirectResponse(redirect_to)

    def locateChild(self, request, segments):
        # URLS CAS
        if segments and segments[0] in ('login',''):
            return self, ()
        if segments and segments[0] == 'verify':
            return AuthenticationVerify(self.manager), ()
        if segments and segments[0] in ['validate','serviceValidate']:
            return CASCompliantResponse(self.manager), ()
        if segments and segments[0] == 'proxyValidate':
            return CASCompliantResponse(self.manager, True), ()
        if segments and segments[0] == 'proxy':
            return ProxyResource(self.manager), ()
        if segments and segments[0] == 'loggedin':
            return LoggedInForm(), ()
        if segments and segments[0] == 'logout':
            return saml_idp.SamlLogout(self.manager), ()
            # return LogoutForm(self.manager), ()
        if segments and segments[0] == 'gen_cookie':
            return LocalCookie(self.manager), ()
        # URLS XMLRPC
        if segments and segments[0] == 'xmlrpc':
            return self.auth_server, ()
        # URLS OPENID
        if segments and segments[0] == 'oid':
            return self.oidserver, segments[1:]
        # URLS SAML
        if segments and segments[0] == 'saml':
            return saml_idp.SamlResponse(self.manager), segments[1:]
        # URLS UTILITAIRES
        if segments and segments[0] == 'css':
            return static.File('interface'), segments[1:]
        if segments[0] == 'images':
            return static.File('interface/images'), segments[1:]
        return ErrorPage(responsecode.NOT_FOUND, u"Ressource non trouvée %s" % '/'.join(segments)), ()

def run():
    if '--test' in sys.argv:
        import doctest
        doctest.testmod(verbose=True)
    else:
        global PARENT_URL
        # intialisation du serveur
        static_data={'rne': [RNE], 'nom_etab': [ETABLISSEMENT]}
        proxy = LDAPProxy(LDAP_SERVER, static_data)
        # lancement du serveur
        if PARENT_URL == '':
            PARENT_URL = None
        log.startLogging(sys.stdout)
        # manager de sessions
        manager = SSOSessionManager(proxy, PARENT_URL)
        # frontend xmlrpc
        auth_server = XMLRPCManager(manager)
        # serveur openid
        baseurl = "https://%s:%s/oid/" % (AUTH_SERVER_ADDR, str(EOLESSO_PORT))
        oidserver = OIDRoot(manager, baseurl)
        # serveur CAS (racine du site)
        root = TestForm(manager, auth_server, oidserver)
        # si le certificat n'existe pas, on le génère
        if not os.path.isfile(CERTFILE):
            log.msg("** Génération d'un nouveau certificat ssl ...")
            gen_certif(CERTFILE)
        log.msg("** Initialisation du serveur SSO")
        reactor.listenSSL(int(EOLESSO_PORT), channel.HTTPFactory(server.Site(root)),
                          ServerContextFactory(CERTFILE))
        reactor.run()

if __name__ == '__main__':
    run()
