#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authserver.py
#
# serveur xmlrpc de gestion de session SSO Eole
#
###########################################################################

import socket
import xmlrpclib
import ldap
from md5 import md5
from glob import glob
from ConfigParser import ConfigParser
# imports twisted
from twisted.python import log
from twisted.internet import reactor
# imports sso eole
from config import *
from saml_utils import get_metadata
from eolesso.ticket import AppTicket
from eolesso.util import *
from eolesso import ticketcache
from eolesso.ticketcache import InvalidSession
# import des fonctions de calcul de données utilisateur
import user_infos

class SSOSessionManager(object):

    def __init__(self,data_proxy, parent=None):
        self.address = socket.getfqdn()
        self.init_caches()
        # proxy d'authentification
        self._data_proxy = data_proxy
        # self._init_proxies()
        self._parent = None
        self.parent_url = parent
        if parent:
            self._parent = xmlrpclib.ServerProxy('%s/xmlrpc' % parent)
            log.msg("\nServeur parent déclaré : %s" % parent)
        self.filters = {}
        self.apps = {}
        self.sp_meta = {}
        self.load_filters()
        # sessions en cours de logout
        self.pending_logout = {}
        # chargement des metadata SAML
        self.init_metadata()

    #def _init_proxies(self):
    #    if not type(LDAP_SERVER) == list:
    #        LDAP_SERVER = [LDAP_SERVER]
    #    if not type(BASE_LDAP) == list:
    #        BASE_LDAP = [BASE_LDAP]
    #    if not type(LDAP_PORT) == list:
    #        LDAP_PORT = [LDAP_PORT]
    #    LDAP_PORT = [int(port) for port in LDAP_PORT]

    #    for ldap_addr in LDAP_SERVER:
    #    # XXX FIXME : gestion multi proxies

    def init_caches(self):
        # cache des "login tickets" (pour empêcher les navigateurs de 'rejouer' l'authentification avec un cache)
        self.login_sessions = {}
        # cache des "granting tickets (ticket de session sso)"
        self.user_sessions = {}
        # cache des attributs utilisateur
        self.data_cache = {}
        # cache des "application" et "proxy" tickets
        self.app_sessions = {}
        # dictionnaire d'association session utilisateur / tickets
        self.user_app_tickets = {}
        # cache pour la gestion des sessions de proxys (expirent en même temps que la session utilisateur)
        self.proxy_granting_sessions = {}
        # cache pour la gestion du single logout (logouts en cours)
        # when single logout requested : we store sso_session and app_session of requester if present
        self.single_logout_requests = {}


    def check_ticket_issuer(self, ticket_id):
        """vérifie qu'un ticket a été généré par ce serveur
        """
        if ticket_id != '':
            ticket_address = "-".join(ticket_id.split('-')[1:-1])
            if ticket_address == self.address:
                return True
        return False

    def end_session(self, session_id):
        """termine la session d'un utilisateur
        """
        # supression des différents tickets associés
        if session_id in self.user_sessions:
            # recherche des tickets associées à la session
            for ticket in self.user_app_tickets[session_id]:
                # recherche des éventuels PGT associés à cet AppTicket
                if ticket.pgt != None:
                    del self.proxy_granting_sessions[ticket.pgt]
                # self.user_app_tickets[session_id].remove(ticket)
                del self.app_sessions[ticket.ticket]
            if session_id in self.user_app_tickets:
                del self.user_app_tickets[session_id]
            del self.user_sessions[session_id]
            log.msg('Fin de la session %s' % session_id)

    def load_applications(self, config):
        """Chargement des applications associées aux filtres
        """
        loaded_apps = []
        for app in config.sections():
            try:
                app_infos = dict(config.items(app))
                try:
                    port = app_infos['port'] = int(app_infos['port'])
                except:
                    port = app_infos['port'] = 'all'
                scheme = app_infos['scheme']
                if scheme not in ['all', 'both']:
                    if scheme not in self.apps:
                        self.apps[scheme]={'all':{}}
                    app_stores = [self.apps[scheme]]
                elif scheme == 'both':
                    app_stores = [self.apps['http'],self.apps['https']]
                else:
                    # pas de protocole spécifié, on stocke dans tous les protocoles
                    app_stores = self.apps.values()
                # on trie les applications par port et chemin
                for store in app_stores:
                    port_store = store.setdefault(port, {})
                    if app_infos['baseurl'] in port_store:
                        log.msg('Warning : baseurl %s already defined for protocol %s (overwritten)' % (app_infos['baseurl'],scheme))
                    port_store[app_infos['baseurl']] = [scheme,
                                                        app_infos['addr'],
                                                        app_infos['typeaddr'],
                                                        app_infos['filter']]
                    loaded_apps.append(app)
            except Exception, e:
                log.msg("! Error loading application %s : %s !" % (app, e))
        log.msg(self.apps)
        return loaded_apps

    def load_filters(self):
        """chargement des filtres de données et applications
        """
        loaded_apps = []
        for conf_file in glob('%s/*.ini' % FILTER_DIR):
            conf_filename = os.path.splitext(os.path.basename(conf_file))[0]
            config = ConfigParser()
            config.read(conf_file)
            if conf_filename.endswith('_apps') or conf_filename == 'applications':
                # liste des applications associées aux filtres
                loaded_apps += self.load_applications(config)
            else:
                # ajout d'un filtre de données
                app_filter = dict( (section, dict(config.items(section)))
                                   for section in config.sections() )
                self.filters[conf_filename] = app_filter
                log.msg("- Filtre chargé: %s" % conf_filename)
        log.msg("- Applications chargées: %s" % ", ".join(loaded_apps))

    def init_metadata(self):
        for conf_file in glob('%s/*.xml' % METADATA_DIR):
            sp_ident = os.path.splitext(os.path.basename(conf_file))[0]
            self.load_metadata(sp_ident)

    def load_metadata(self, sp_ident):
        try:
            sp_meta = get_metadata(sp_ident)
            if 'entityID' in sp_meta:
                sp_ident = sp_meta['entityID']
        except Exception, e:
            log.msg("- Erreur au chargement du fichier de metadata %s (%s)" % (os.path.basename(conf_file), str(e)))
            return {}
        self.sp_meta[sp_ident] = sp_meta
        log.msg("- Fournisseur d'identité configuré : %s" % sp_ident)
        return sp_meta

    def get_metadata(self, sp_ident):
        if sp_ident in self.sp_meta:
            return self.sp_meta[sp_ident]
        else:
            try:
                sp_meta = self.load_metadata(sp_ident)
            except:
                sp_meta = {}
        return sp_meta

    def authenticate(self, username, password):
        sessionID = ''
        if type(username) == unicode: username = username.encode(encoding)
        if type(password) == unicode: password = password.encode(encoding)

        # on demande en premier au serveur père qu'il vérifie les informations
        if self._parent is not None:
            sessionID = self._parent.authenticate(username, password)
        if sessionID != '':
            log.msg('Session délivrée par le serveur parent (%s)' % username)
            # session du parent, on crée une session locale 'déportée'
            # on ne gère pas d'expiration sur cette session, la validité sera vérifiée sur le parent
            self.user_sessions[sessionID] = None
            self.user_app_tickets[sessionID] = []
        elif password != '':
            # non authentifié par un parent, on essaye de le faire localement
            try:
                authenticated, user_data = self._data_proxy.authenticate(username, password)
                if authenticated:
                    sessionID = gen_ticket_id('TGC', self.address)
                    if sessionID != "":
                        self.user_sessions[sessionID] = username
                        # stockage des données utilisateur
                        self.data_cache[sessionID] = user_data
                        # initialisation de la liste des tickets associés
                        self.user_app_tickets[sessionID] = []
                        reactor.callLater(SESSION_TIMER, self.end_session, sessionID)
                        log.msg("Session créée (%s)" % username)
                    return sessionID
            except ldap.SERVER_DOWN:
                log.msg("! Erreur d'accès au serveur LDAP !")
                return ''
            log.msg("! Echec de l'authentification : %s !" % username)
            return ''
        return sessionID

    def get_login_ticket(self):
        lt = gen_ticket_id('LT', self.address)
        self.login_sessions[lt] = 'login'

    def validate_session(self, session_id):
        """vérifie un session_id existant"""
        if self.check_ticket_issuer(session_id) and session_id in self.user_sessions:
            return True
        if self._parent is not None:
            valid_session = self._parent.validate_session(session_id)
            if valid_session:
                log.msg('Session utilisateur validée par le serveur parent')
                # si la session n'est pas présente localement, mais validée par le parent, on l'ajoute (ex : redémarrage du serveur local)
                if not session_id in self.user_sessions:
                    self.user_sessions[session_id] = None
                    self.user_app_tickets[session_id] = []
            return valid_session
        log.msg("! Session invalide : %s !" % session_id)
        return False

    def verify_session_id(self, session_id):
        """vérifie un session_id"""
        if self.check_ticket_issuer(session_id) and session_id in self.user_sessions:
                return True, session_id
        if self._parent is not None:
            return self._parent.verify_session_id(session_id)
        log.msg("! session utilisateur non reconnue : %s !" % session_id)
        return False, session_id

    def get_user_info(self, session_id, details=False):
        """méthode de convenance pour récupérer l'utilisateur"""
        if session_id in self.user_sessions:
            if self.check_ticket_issuer(session_id):
                if details:
                    return self._get_user_details(session_id, "", {})
                else:
                    return self.user_sessions[session_id]
            else:
                # session du serveur parent
                return self._parent.get_user_info(session_id, details)
        raise InvalidSession(session_id)

    def logout(self, session_id):
        """vérifie un session_id"""
        if session_id in self.user_sessions:
            self.end_session(session_id)
            if not self.check_ticket_issuer(session_id):
                # session délivrée par un le parent
                return self._parent.logout(session_id)
            else:
                return 'ok'
        else:
            return ''

    def get_app_ticket(self, session_id, appurl, ticket_prefix='ST', from_credentials=False):
        # si cette session utilisateur est valide, on crée un ticket pour l'application
        appSessionID = ''
        if self.validate_session(session_id):
            # génération d'un nouveau ticket d'application associé à la session
            appSessionID = gen_ticket_id(ticket_prefix, self.address)
            ticket = self.init_app_session(appSessionID, session_id, appurl, from_credentials)
            if ticket.filter != '':
                filter_msg = ' (filtre de données : %s)' % ticket.filter
            else:
                filter_msg = ''
            log.msg("Session autorisée pour l'application %s%s" % (ticket.service_url, filter_msg))
            return appSessionID
        else:
            log.msg("Session utilisateur non reconnue !")
        log.msg("! Echec de création de session pour l'application %s :" % appurl)
        return ''

    def init_app_session(self, appSessionID, session_id, appurl, from_credentials):
        """génère un ticket d'application et initialise son timeout
        """
        ticket = AppTicket(appSessionID, session_id, appurl, self.address, from_credentials=from_credentials)
        # détection d'un éventuel filtre sur les données utilisateur
        ticket.filter = self._check_filter(ticket.service_url)
        log.msg("adding session for service  : %s" % ticket.service_url)
        # stockage du ticket
        self.app_sessions[appSessionID] = ticket
        self.user_app_tickets[session_id].append(ticket)
        # mise en place des timeouts de session
        reactor.callLater(APP_TIMER, self.invalidate_app_ticket, appSessionID)
        return ticket

    def invalidate_app_ticket(self, appSessionID):
        """invalide un ticket d'application pour empêcher son utilisation
        """
        if appSessionID in self.app_sessions:
            self.app_sessions[appSessionID].valid = False

    def verify_app_ticket(self, app_ticket, appurl):
        """vérifie le ticket fourni par une application"""
        verified = False
        ticket = app_ticket
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            verified = ticket.verif_ticket(app_ticket, appurl)
        else:
            # ticket inconnu, essai de vérification sur un serveur parent
            if self._parent is not None:
                return self._parent.verify_app_ticket(app_ticket, appurl)
        if not verified:
            log.msg("! Echec de la vérification de session pour l'application %s !" % get_service_from_url(appurl))
        return verified, app_ticket

    def get_user_details(self, app_ticket, appurl, sections=False, renew=False, keep_valid=False):
        """vérifie un ticket et renvoie les informations sur l'utilisateur
        section : si True, on renvoie aussi une description de l'organisation des données
        """
        log.msg('--- validating session for %s' % (appurl))
        code = 'INVALID_TICKET'
        detail = """Le ticket %s n'est pas reconnu par le serveur SSO""" % app_ticket
        if app_ticket in self.app_sessions:
            log.msg('--- session ok')
            ticket = self.app_sessions[app_ticket]
            # si le paramètre renew a été envoyé et que le ticket provient d'une
            # session SSO, on ne le valide pas
            if not ticket.from_credentials and renew:
                detail = """Le ticket %s n'a pas été délivré lors de la phase d'authentification""" % app_ticket
            elif ticket.verif_ticket(app_ticket, appurl, keep_valid):
                log.msg('--- appurl verified')
                return True, self._get_user_details(ticket.session_id, ticket.filter, sections)
            else:
                code = 'INVALID_SERVICE'
                detail = """Ce ticket n'appartient pas au service %s""" % appurl
                log.msg("! Lecture des données utilisateur : session invalide !")
        else:
            # session inconnue sur ce serveur, on essaie sur le serveur parent
            if self._parent is not None:
                # récupération des infos sur le serveur parent
                # XXX  FIXME : re-filtrer les infos du serveur parent avec les filtres locaux ??
                return self._parent.get_user_details(app_ticket, appurl, sections, renew)
        log.msg("! Lecture des données utilisateur refusé pour l'application %s !" % appurl)
        if sections:
            return False, ({'code':code,'detail':detail},"")
        return False, {'code':code,'detail':detail}

    def get_proxy_granting_ticket(self, app_ticket, service, pgturl):
        # si une url de proxy est donnée, on crée un ticket TGC
        # la conformité de l'url est vérifiée lors de la création.
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            if ticket.verif_ticket(app_ticket, service, keep_valid=True):
                ticket.generate_pgt(pgturl)
                self.proxy_granting_sessions[ticket.pgt] = ticket
                return ticket
        return None

    def get_proxy_ticket(self, pgt, target_service):
        # récupération de la session
        app_ticket = self.proxy_granting_sessions[pgt]
        appSessionID = gen_ticket_id('PT', self.address)
        # cas spécial : imapproxy ne fonctionne pas avec des pwd de + de 61 caractères
        if target_service.startswith('imap://') or target_service.startswith('imaps://'):
            appSessionID = appSessionID[:60]
        ticket = self.init_app_session(appSessionID, app_ticket.session_id, target_service, app_ticket.from_credentials)
        ticket.parent = app_ticket
        return ticket.ticket

    def invalidate_proxy(self, app_ticket):
        """Supprime les paramètres de proxy d'un app_ticket si les vérifications sur l'url de callback ont échoué
        """
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            if ticket.pgt is not None:
                del self.proxy_granting_sessions[ticket.pgt]
            ticket.reset_pgt()

    def _check_filter(self, appurl):
        """vérifie si un filtre est disponible pour l'url"""
        url = urlparse.urlparse(appurl)
        if url.scheme in self.apps:
            store = self.apps[url.scheme]
        else:
            store = self.apps['http']
        # recherche d'une application avec le même port/chemin
        if url.port in store:
            apps = store[url.port]
        else:
            # si pas de port correspondant, on regarde dans les applis sans port spécifique
            apps = store['all']
        # gestion du path, il faut que le path commence par ce qu'on veut
        # on trie les chemins connus pour commencer par le plus spécifique
        app_paths = apps.keys()
        app_paths.sort(reverse=True)
        url_path = url.path
        if not url_path.startswith('/'): url_path = '/' + url_path
        for app in app_paths:
            if not app.startswith('/'): app = '/' + app
            if url_path.startswith(app):
                scheme, addr, typeaddr, id_filter = apps[app]
                if typeaddr == 'ip':
                    addr_ok = check_hostname_by_ip(url, addr)
                else:
                    addr_ok = check_hostname_by_domain(url, addr)
                if addr_ok:
                    # on a une appli correspondant à l'url, on renvoie le filtre associé
                    log.msg("Filtre %s appliqué pour l'url %s" % (id_filter, appurl))
                    return id_filter
                break

        return "default"


    def _filter_data(self, infos, id_filter, sections):
        """filtrage en fonction de l'application si des filtres sont définis"""
        data = {}
        if id_filter in self.filters:
            for section, vars in self.filters[id_filter].items():
                for libelle_cas, nom_val in vars.items():
                    data[nom_val] = infos.get(nom_val, '')
            filter_data = self.filters[id_filter]
        else:
            # FIXME : définir les données à renvoyer par défaut !
            data.update(infos)
            filter_data = ""
        if sections:
            return data, filter_data
        else:
            return data

    def _get_user_details(self, session_id, id_filter, sections):
        """Renvoie les informations connues de l'utilisateur lié à <session_id>"""
        # on récupère l'id de l'utilisateur
        user_id = self.user_sessions[session_id]
        if self.check_ticket_issuer(session_id):
            infos = self.data_cache[session_id]
        else:
            # session distante
            infos = self._parent.get_user_info(session_id,True)
        # ajout des informations statiques
        infos.update({'rne':[RNE],'nom_etab':[ETABLISSEMENT]})
        # lancement des fonctions de calcul de valeurs supplémentaires
        for varname, calc_func in user_infos.dict_infos.items():
            try:
                infos[varname] = calc_func(infos)
            except Exception, e:
                log.msg("""erreur lors du calcul du champ '%s' :  %s""" % (varname, str(e)))
        infos = self._filter_data(infos, id_filter, sections)
        log.msg('--- user data sent')
        return infos


def run(*args):
    global PARENT_URL
    # intialisation du serveur
    from eolesso.dataproxy import LDAPProxy
    from twisted.web2 import server
    from eolesso.libsecure import ServerContextFactory
    proxy = LDAPProxy(LDAP_SERVER)
    # lancement du serveur
    if PARENT_URL == '':
        PARENT_URL = None
    reactor.listenSSL(int(AUTH_SERVER_PORT), server.Site(SSOSessionManager(proxy,PARENT_URL)),ServerContextFactory(CERTFILE))
    print "\nServeur initialisé (port %s)\n" % AUTH_SERVER_PORT
    reactor.run()


if __name__ == '__main__':
    run()
