# -*- coding: utf-8 -*-
###########################################################################
# Eole NG - 2007  
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr 
#  
# config.py
#  
# fichier de configuration pour la librairie SSO Eole
#       
###########################################################################

import locale, os, socket
from creole.parsedico import parse_dico
from creole.config import cert_file

# dico local pour Eole1
dico = parse_dico()
ip_locale = dico['adresse_ip_eth0']
# configuration du serveur ldap
if dico.has_key('base_ldap_sso'):
    base_ldap = dico['base_ldap_sso']
else:
    base_ldap = "o=gouv,c=fr"
if dico.has_key('ldap_sso'):
    serveur_ldap = dico['ldap_sso']
elif dico.has_key('serveur_ldap'):
    serveur_ldap = dico['serveur_ldap']
else:
    serveur_ldap = dico['adresse_ip_eth0']
# identifiant rne du serveur si disponible
if dico.has_key('numero_etab'):
    RNE=dico['numero_etab']
else:
    RNE=""
if dico.has_key('libelle_etab'):
    ETABLISSEMENT=dico['libelle_etab']
else:
    ETABLISSEMENT=""
# durée de vie d'un login ticket : 2 heures
try:
    assert dico.has_key('sso_session_timeout')
    SESSION_TIMER = int(dico['sso_session_timeout'])
except:
    SESSION_TIMER = 2 * 60 * 60
# durée de vie d'un ticket d'application : 5 minutes (utilisable une seule fois)
APP_TIMER = 300
# durée de vie d'un nombre magique permettant la communication entre un frontend et un backend
# ce timeout est reinitialise a chaque fois qu'une action est lancée
MAGIC_NUMBER_TIMEOUT =  60 * 5
# durée de vie des sessions expirées. Quand une session expire, on la garde dans un cache pour savoir
# si quand une session est non valide, si elle avait expirée ou si elle n'a jamais existée
EXPIRED_SESSION_TIMEOUT = 60 * 60 * 2
# infos sur le serveur ldap
LDAP_SERVER = serveur_ldap
if dico.has_key('base_ldap_sso'):
    LDAP_BASE = dico['base_ldap_sso']
else:
    LDAP_BASE = base_ldap
try:
    assert dico.has_key('port_ldap_sso')
    LDAP_PORT = int(dico['port_ldap_sso'])
except:
    LDAP_PORT = 389
# adresse du serveur d'authentification
AUTH_SERVER_ADDR = socket.getfqdn(ip_locale)
# port de la page web d'authentification
if dico.has_key('port_sso'):
    EOLESSO_PORT = dico['port_sso']
else:
    EOLESSO_PORT = '8443'
# version de CAS (1 ou 2)
try:
    assert dico.has_key('cas_version')
    CAS_VERSION = int(dico['cas_version'])
except:
    CAS_VERSION = 2
# URL de la page web d'authentification
AUTH_FORM_URL = 'https://%s:%s' % (AUTH_SERVER_ADDR, EOLESSO_PORT)
# URL du serveur d'authentification
AUTH_SERVER='https://%s:%s/xmlrpc' % (AUTH_SERVER_ADDR, EOLESSO_PORT)
# URL du serveur parent si déclaré
PARENT_URL = ''
if dico.has_key('adresse_sso_parent'):
    if dico['adresse_sso_parent'] != '':
        PARENT_URL = 'https://%s:%s' % (dico['adresse_sso_parent'],dico['port_sso_parent']) 
# certificats ssl du serveur sso
if dico.has_key('sso_cert') and os.path.isfile(dico['sso_cert']):
    CERTFILE = dico['sso_cert']
else:
    CERTFILE = cert_file
# certificat de l'autorité de certification (pour validation des clients en mode proxy)
if dico.has_key('sso_ca_location') and os.path.isfile(dico['sso_ca_location']):
    CA_LOCATION = dico['sso_ca_cert']
else:
    CA_LOCATION = os.path.join(os.path.dirname(cert_file), 'ca.crt')
# répertoire des filtres d'application
FILTER_DIR = os.path.join(os.path.dirname(__file__),'app_filters')
# répertoire des fichiers metadata pour les entités SAML
METADATA_DIR = os.path.join(os.path.dirname(__file__),'metadata')
NAMEID_FIELD = 'mail'
# SAML IDP entity name
IDP_IDENTITY = "urn:fi:ac-dijon:et-%s:1.0" % (ETABLISSEMENT or AUTH_SERVER_ADDR)
# encoding par défaut
encoding = locale.getpreferredencoding()
